/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                           BIAYSTOK 2018                               **
**                                                                        **
**                         Home Control System                            **
**                                                                        **
**                               HC_SCR1                                  **
**                      Wersja z dnia : 2018.06.21                        **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include "scr1eeprom.h"



#define false                          0
#define true                           1

#define LEDPort                        PORTC
#define LEDPortDirection               DDRC
#define StatusLed                      0
#define RxTxLed                        1
#define ConnectionLed                  2

#define SegmPort                       PORTA
#define SegmDirPort                    DDRA
#define DigitPort                      PORTB
#define DigitDirPort                   DDRB
#define Digit0                         3
#define Digit1                         2
#define Digit2                         1
#define Digit3                         0

#define SegmentA                       0x01
#define SegmentB                       0x02
#define SegmentC                       0x04
#define SegmentD                       0x08
#define SegmentE                       0x10
#define SegmentF                       0x20
#define SegmentG                       0x40
#define SegmentK                       0x80

#define TimerCounterDelayLimit         1024 // 2048
#define StandardLEDPulseDelay          832

#define RxTxDirPort                    PORTD
#define RxTxDirDirectionPort           DDRD
#define RxTxDirPin                     4              // PORTD . 4

#define null                           0
#define cr                             0x0D
#define Space                          0x20
#define PartSeparator                  ':'
#define ParamSeparator                 '.'

#define nop() __asm__ __volatile__ ("nop")

// UART BAUD Rate Generator fosc=7.3728 MHz
#define Speed9600bps                   47
#define Speed19200bps                  23
#define Speed38400bps                  11
#define Speed115200bps                 3

#define IdleState                      0
#define BusyState                      1

#define TXDBuffSize                    64
#define RXDBuffSize                    16

#define CyclBuffSize                   64

typedef struct { uint8_t CycFr ;
                 uint8_t CycTo ;
                 uint8_t Buffer [ CyclBuffSize ] ;
               } CyclicRecT ;

typedef struct { volatile uint8_t CycFr ;
                 volatile uint8_t CycTo ;
                 volatile uint8_t State ;
                 uint8_t Buffer [ TXDBuffSize ] ;
               } TXDRecT ;
typedef struct { volatile uint8_t CycFr ;
                 volatile uint8_t CycTo ;
                 uint8_t Buffer [ RXDBuffSize ] ;
               } RXDRecT ;

#define CommandBufferSize              16
typedef uint8_t CommandBufferType [ CommandBufferSize + 1 ] ;
#define ParamBufferSize                32
typedef uint8_t ParamBufferType [ ParamBufferSize + 1 ] ;

struct _SideServiceStruct ;

typedef struct _SideServiceStruct * PtrSideServiceType ;

typedef void ( * ParserOperationProcT ) ( void ) ;

typedef struct _SideServiceStruct {
                 TXDRecT                   TXDRec ;
                 RXDRecT                   RXDRec ;
                 ParserOperationProcT      ParserService ;
                 CommandBufferType         CommandText ;
                 ParamBufferType           ParamText ;
                 DeviceIDType              DestinationDeviceName ;
                 DeviceIDType              SourceDeviceName ;
                 uint8_t                   CommandTextIndex ;
                 uint8_t                   ParamTextIndex ;
                 uint8_t                   DestDeviceNameIndex ;
                 uint8_t                   SrcDeviceNameIndex ;
                 uint16_t                  SerialLEDDelay ;
               } SideServiceType ;


static CyclicRecT CyclicBuffer ;
static SideServiceType UpService ;
static uint8_t ConnectContext ;
static uint8_t CurrentTimeStatusState ;
static uint16_t TimerCounter ;
static uint8_t DigitCounter ;
volatile static uint8_t DispBuffer [ 4 ] ;
volatile static uint8_t TimeEvent ;

static const uint16_t CRCTable [ 16 ] = { 0x0000 ,  /* 00 */
                                          0xCC01 ,  /* 01 */
                                          0xD801 ,  /* 02 */
                                          0x1400 ,  /* 03 */
                                          0xF001 ,  /* 04 */
                                          0x3C00 ,  /* 05 */
                                          0x2800 ,  /* 06 */
                                          0xE401 ,  /* 07 */
                                          0xA001 ,  /* 08 */
                                          0x6C00 ,  /* 09 */
                                          0x7800 ,  /* 10 */
                                          0xB401 ,  /* 11 */
                                          0x5000 ,  /* 12 */
                                          0x9C01 ,  /* 13 */
                                          0x8801 ,  /* 14 */
                                          0x4400 } ;/* 15 */

static uint8_t MyDefaultID [ ] PROGMEM              = "SCR1" ;
static uint8_t DeviceName [ ] PROGMEM               = "HCSCR1-wyswietlacz liczbowy\r" ;
static uint8_t HardwareSoftwareVersion [ ] PROGMEM  = "HVER*1_00*V-2011.SVER*1_0*VI-2018\r" ;
static uint8_t OK [ ] PROGMEM                       = "<OK>.\r" ;
static uint8_t Fail [ ] PROGMEM                     = "<FAIL>.\r" ;
static uint8_t GeneralPoolDeviceName [ ] PROGMEM    = "GENERAL" ;
static uint8_t NoDataText [ ] PROGMEM                = "NODATA\r" ;

static uint8_t HelloCommand [ ] PROGMEM             = "HELLO" ;
static uint8_t ConnectCommand [ ] PROGMEM           = "CONNECT" ;
static uint8_t DisConnectCommand [ ] PROGMEM        = "DISCONNECT" ;
static uint8_t SetDeviceNameCommand [ ] PROGMEM     = "SETMID" ;
static uint8_t GetVersionCommand [ ] PROGMEM        = "GETVER" ;
static uint8_t GetIDCommand [ ] PROGMEM             = "GETID" ;
static uint8_t ReinitCommand [ ] PROGMEM            = "REINIT" ;
static uint8_t ResetEECommand [ ] PROGMEM           = "RESEE" ;
static uint8_t SaveEECommand [ ] PROGMEM            = "SAVEEE" ;
static uint8_t NewPasswordCommand [ ] PROGMEM       = "NEWPW" ;
static uint8_t ReplyCommand [ ] PROGMEM             = "REPLY" ;
static uint8_t DispCommand [ ] PROGMEM              = "DA" ;
static uint8_t DispBinCommand [ ] PROGMEM           = "DS" ;


typedef void ( * CommandExecuteServiceProcT ) ( void ) ;
typedef struct {
                 CommandExecuteServiceProcT  CommandService ;
                 uint8_t *                   CommandName ;
               } CommandNameAssociationRecT ;

static void ConnectCommandExecute ( void ) ;
static void DisconnectCommandExecute ( void ) ;
static void HelloCommandExecute ( void ) ;
static void GetVersionCommandExecute ( void ) ;
static void GetIDCommandExecute ( void ) ;
static void ReinitCommandExecute ( void ) ;
static void ResetEECommandExecute ( void ) ;
static void SaveEECommandExecute ( void ) ;
static void SetDeviceNameCommandExecute ( void ) ;
static void NewPasswordCommandExecute ( void ) ;
static void ReplyCommandExecute ( void ) ;
static void DispCommandExecute ( void ) ;
static void DispBinCommandExecute ( void ) ;


#define CommandNameAssociationTabeSize             13

static CommandNameAssociationRecT CommandNameAssociationTabe [ CommandNameAssociationTabeSize ] PROGMEM =
{
/* 00 */ { ConnectCommandExecute               , ConnectCommand } ,
/* 01 */ { DisconnectCommandExecute            , DisConnectCommand } ,
/* 02 */ { SetDeviceNameCommandExecute         , SetDeviceNameCommand } ,
/* 03 */ { HelloCommandExecute                 , HelloCommand } ,
/* 04 */ { GetVersionCommandExecute            , GetVersionCommand } ,
/* 05 */ { GetIDCommandExecute                 , GetIDCommand } ,
/* 06 */ { ReinitCommandExecute                , ReinitCommand } ,
/* 07 */ { ResetEECommandExecute               , ResetEECommand } ,
/* 08 */ { SaveEECommandExecute                , SaveEECommand } ,
/* 09 */ { NewPasswordCommandExecute           , NewPasswordCommand } ,
/* 10 */ { ReplyCommandExecute                 , ReplyCommand } ,
/* 11 */ { DispCommandExecute                  , DispCommand  } ,
/* 12 */ { DispBinCommandExecute               , DispBinCommand } ,
} ;



static uint8_t HexalTable [ ] PROGMEM                = "0123456789ABCDEF" ;

typedef enum { LifeStatusLED ,
               SuperVRxTxLED ,
               ConnectLED ,
             } LEDSinalT ;




  /*----------------------------------------------------------------*/
static void LEDOn ( LEDSinalT LEDSignal )
{
  switch ( LEDSignal )
  {
    case LifeStatusLED :
      LEDPort &= ~ ( 1 << StatusLed ) ;
      break ;
    case SuperVRxTxLED :
      LEDPort &= ~ ( 1 << RxTxLed ) ;
      break ;
    case ConnectLED :
      LEDPort &= ~ ( 1 << ConnectionLed ) ;
      break ;
  } /* switch */ ;
} /* LEDOn */


static void LEDOff ( LEDSinalT LEDSignal )
{
  switch ( LEDSignal )
  {
    case LifeStatusLED :
      LEDPort |= ( 1 << StatusLed ) ;
      break ;
    case SuperVRxTxLED :
      LEDPort |= ( 1 << RxTxLed ) ;
      break ;
    case ConnectLED :
      LEDPort |= ( 1 << ConnectionLed ) ;
      break ;
  } /* switch */ ;
} /* LEDOff */


static void DisplayRfsh ( void )
{
  uint8_t Bits ;
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  Bits = ( 1 << Digit0 ) | ( 1 << Digit1 ) | ( 1 << Digit2 ) | ( 1 << Digit3 ) ;
  DigitPort |= Bits ;
  DigitCounter ++ ;
  DigitCounter &= 0x03 ;
  switch ( DigitCounter )
  {
    case 0 :
      Bits = ~ ( 1 << Digit0 ) ;
      Data = DispBuffer [ 3 ] ;
      break ;
    case 1 :
      Bits = ~ ( 1 << Digit1 ) ;
      Data = DispBuffer [ 2 ] ;
      break ;
    case 2 :
      Bits = ~ ( 1 << Digit2 ) ;
      Data = DispBuffer [ 1 ] ;
      break ;
    case 3 :
      Bits = ~ ( 1 << Digit3 ) ;
      Data = DispBuffer [ 0 ] ;
      break ;
  } /* switch */ ;
  SegmPort = Data ;
  DigitPort &= Bits ;
} /* DisplayRfsh */


ISR ( TIMER0_OVF_vect )
{
  TimerCounter ++ ;
  if ( TimerCounter > TimerCounterDelayLimit )
  {
    TimerCounter = 0 ;
    TimeEvent = true ;
  } /* if */ ;
  if ( UpService . SerialLEDDelay )
  {
    UpService . SerialLEDDelay -- ;
    if ( ! UpService . SerialLEDDelay )
    {
      LEDOff ( SuperVRxTxLED ) ;
    } /* if */ ;
  } /* if */ ;
  DisplayRfsh ( ) ;
} /* TIMER0_OVF_vect */


static void StartSvRxTxLED ( void )
{
  LEDOn ( SuperVRxTxLED ) ;
  UpService . SerialLEDDelay = StandardLEDPulseDelay ;
} /* StartSvRxTxLED */


static void UARTReceiveMode ( void )
{
  RxTxDirPort &= ~ ( 1 << RxTxDirPin ) ;
} /* UARTReceiveMode */


static void UARTTransmitMode ( void )
{
  RxTxDirPort |= ( 1 << RxTxDirPin ) ;
} /* UARTTransmitMode */


ISR ( UART_RX_vect )
{
  uint8_t RecCh ;
  /*----------------------------------------------------------------*/
  RecCh = UDR ;
  UpService . RXDRec . Buffer [ UpService . RXDRec . CycTo ] = RecCh ;
  UpService . RXDRec . CycTo ++ ;
  if ( UpService . RXDRec . CycTo >= RXDBuffSize )
    UpService . RXDRec . CycTo = 0 ;
  StartSvRxTxLED ( ) ;
} /* UART_RX_vect */


ISR ( UART_TX_vect )
{
  if ( UpService . TXDRec . CycFr == UpService . TXDRec . CycTo )
  {
    UpService . TXDRec . State = IdleState ;
    UARTReceiveMode ( ) ;
  } /* if ... */
  else
  {
    UDR = UpService . TXDRec . Buffer [ UpService . TXDRec . CycFr ] ;
    UpService . TXDRec . CycFr ++ ;
    if ( UpService . TXDRec . CycFr >= TXDBuffSize )
      UpService . TXDRec . CycFr = 0 ;
    StartSvRxTxLED ( ) ;
  } /* if ... else */ ;
} /* UART_TX_vect */


static void SendSerial ( uint8_t SerialCh )
{
  uint8_t NextInx ;
  /*----------------------------------------------------------------*/
  UARTTransmitMode ( ) ;
  StartSvRxTxLED ( ) ;
  cli ( ) ;
  if ( UpService . TXDRec . State == IdleState )
  {
    UpService . TXDRec . State = BusyState ;
    UDR = SerialCh ;
    sei ( ) ;
  } /* if ... */
  else
  {
    for ( ; ; )
    {
      NextInx = UpService . TXDRec . CycTo + 1 ;
      if ( NextInx >= TXDBuffSize )
        NextInx = 0 ;
      if ( NextInx != UpService . TXDRec . CycFr )
        break ;
      sei ( ) ;
      nop ( ) ;
      nop ( ) ;
      wdt_reset ( ) ;
      cli ( ) ;
    } /* for */ ;
    UpService . TXDRec . Buffer [ UpService . TXDRec . CycTo ] = SerialCh ;
    UpService . TXDRec . CycTo ++ ;
    if ( UpService . TXDRec . CycTo >= TXDBuffSize )
      UpService . TXDRec . CycTo = 0 ;
    sei ( ) ;
  } /* if ... else */ ;
} /* SendSerial */


static uint8_t SerialDataPresent ( void )
{
  if ( UpService . RXDRec . CycFr == UpService . RXDRec . CycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* SerialDataPresent */


static uint8_t GetSerial ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = UpService . RXDRec . Buffer [ UpService . RXDRec . CycFr ] ;
  UpService . RXDRec . CycFr ++ ;
  if ( UpService . RXDRec . CycFr >= RXDBuffSize )
    UpService . RXDRec . CycFr = 0 ;
  return ( Ch ) ;
} /* GetSerial */


static void AddToCyclic ( uint8_t Ch )
{
  CyclicBuffer . Buffer [ CyclicBuffer . CycTo ] = Ch ;
  CyclicBuffer . CycTo ++ ;
  if ( CyclicBuffer . CycTo >= CyclBuffSize )
    CyclicBuffer. CycTo = 0 ;
} /* AddToCyclic */


static void AddToCyclicTxt ( uint8_t * String )
{
  uint8_t Data ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Data = * String ++ ;
    if ( Data )
      AddToCyclic ( Data ) ;
    else
      break ;
  } /* for */ ;
} /* AddToCyclicTxt */


static void AddToCyclicFlashTxt ( uint16_t String )
{
  uint8_t Data ;
  /*------------------------------*/
  for ( ; ; )
  {
    Data = pgm_read_byte_near ( String ) ;
    if ( Data )
      AddToCyclic ( Data ) ;
    else
      break ;
    String ++ ;
  } /* for */ ;
} /* AddToCyclicFlashTxt */


static uint8_t CyclicDataPresent ( void )
{
  if ( CyclicBuffer . CycFr == CyclicBuffer . CycTo )
    return ( false ) ;
  else
    return ( true ) ;
} /* CyclicDataPresent */


static uint8_t GetFromCyclic ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = CyclicBuffer . Buffer [ CyclicBuffer . CycFr ] ;
  CyclicBuffer . CycFr ++ ;
  if ( CyclicBuffer . CycFr >= CyclBuffSize )
    CyclicBuffer . CycFr = 0 ;
  return ( Ch ) ;
} /* GetFromCyclic */


static void FlushCyclic ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  while ( CyclicDataPresent ( ) )
  {
    Ch = GetFromCyclic ( );
    SendSerial ( Ch ) ;
  } /* while */ ;
} /* FlushCyclic */


static void UARTHardwareInit ( uint16_t SvSerialSpeed )
{
  UBRRH = 0 ;
  UBRRL = SvSerialSpeed ;
  UCSRB = ( 1 << RXCIE ) | ( 1 << TXCIE ) | ( 1 << RXEN ) | ( 1 << TXEN ) ;
  UCSRC = ( 1 << URSEL ) | ( 1 << UCSZ1 ) | ( 1 << UCSZ0 ) ;
} /* UARTHardwareInit */



static void HardwareInit ( uint16_t SvSerialSpeed )
{
  TCCR0 = ( 1 << CS01 ) ; // CS02=0 CS01=1 CS00=0 : F clk div 8
  TIMSK = ( 1 << TOIE0 ) ;
  UARTHardwareInit ( SvSerialSpeed ) ;
} /* HardwareInit */


static void EnvirInit ( void )
{
  uint8_t Bits ;
  /*----------------------------------------------------------------*/
  Bits = 0xFF ;
  SegmDirPort = Bits ;
  SegmPort = 0 ;
  Bits = ( 1 << Digit0 ) | ( 1 << Digit1 ) | ( 1 << Digit2 ) | ( 1 << Digit3 ) ;
  DigitDirPort = Bits ;
  DigitPort = Bits ;
  Bits = ( 1 << StatusLed ) | ( 1 << RxTxLed ) | ( 1 << ConnectionLed ) ;
  LEDPortDirection |= Bits ;
  LEDPort |= Bits ;
  Bits = ( 1 << RxTxDirPin ) ;
  RxTxDirDirectionPort |= Bits ;
  UARTReceiveMode ( ) ;
} /* EnvirInit */


static void ClearBlock ( void *   Destination ,
                         uint16_t Size )
{
  uint16_t Loop ;
  uint8_t * Dst ;
  /*----------------------------------------------------------------*/
  Dst = ( uint8_t * ) Destination ;
  for ( Loop = 0 ; Loop < Size ; Loop ++ )
    * Dst ++ = null ;
} /* ClearBlock */


static void StrCopy ( uint8_t * Destination ,
                      uint8_t * Source )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  Ch = * Source ;
  * Destination = Ch ;
  while ( Ch )
  {
    Ch = * Source ++ ;
    * Destination ++ = Ch ;
  } /* while */ ;
} /* StrCopy */


static uint8_t ConvToSegments ( uint8_t Ch )
{
  uint8_t Ret ;
  /*----------------------------------------------------------------*/
  Ret = 0 ;
  switch ( Ch )
  {
    case '-' :
      Ret = SegmentG ;
      break ;
    case '0' :
      Ret = SegmentA | SegmentB | SegmentC | SegmentD | SegmentE | SegmentF ;
      break ;
    case '1' :
      Ret = SegmentB | SegmentC ;
      break ;
    case '2' :
      Ret = SegmentA | SegmentB | SegmentD | SegmentE | SegmentG ;
      break ;
    case '3' :
      Ret = SegmentA | SegmentB | SegmentC | SegmentD | SegmentG ;
      break ;
    case '4' :
      Ret = SegmentB | SegmentC | SegmentF | SegmentG ;
      break ;
    case '5' :
      Ret = SegmentA | SegmentC | SegmentD | SegmentF | SegmentG ;
      break ;
    case '6' :
      Ret = SegmentA | SegmentC | SegmentD | SegmentE | SegmentF | SegmentG ;
      break ;
    case '7' :
      Ret = SegmentA | SegmentB | SegmentC ;
      break ;
    case '8' :
      Ret = SegmentA | SegmentB | SegmentC | SegmentD | SegmentE | SegmentF | SegmentG ;
      break ;
    case '9' :
      Ret = SegmentA | SegmentB | SegmentC | SegmentD | SegmentF | SegmentG ;
      break ;
    case 'a' :
    case 'A' :
      Ret = SegmentA | SegmentB | SegmentC | SegmentE | SegmentF | SegmentG ;
      break ;
    case 'b' :
    case 'B' :
      Ret = SegmentC | SegmentD | SegmentE | SegmentF | SegmentG ;
      break ;
    case 'c' :
    case 'C' :
      Ret = SegmentA | SegmentD | SegmentE | SegmentF ;
      break ;
    case 'd' :
    case 'D' :
      Ret = SegmentB | SegmentC | SegmentD | SegmentE | SegmentG ;
      break ;
    case 'e' :
    case 'E' :
      Ret = SegmentA | SegmentD | SegmentE | SegmentF | SegmentG ;
      break ;
    case 'f' :
    case 'F' :
      Ret = SegmentA | SegmentE | SegmentF | SegmentG ;
      break ;
  } /* switch */ ;
  return Ret ;
} /* ConvToSegments */


static void CodePassword ( void )
{
  uint8_t Loop ;
  uint8_t Element ;
  uint8_t BitCorrection ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Element = EEBlock . Password [ Loop ] ^ 0xFF ;
    BitCorrection = Element & 0x01 ;
    Element = Element >> 1 ;
    if ( BitCorrection )
      Element |= 0x80 ;
    EEBlock . Password [ Loop ] = Element ;
  } /* for */ ;
} /* CodePassword */


static void DecodePassword ( void )
{
  uint8_t Loop ;
  uint16_t Element ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Element = ( uint16_t ) EEBlock . Password [ Loop ] ;
    Element = Element << 1 ;
    if ( Element >= 0x100 )
      Element ++ ;
    Element &= 0xFF ;
    Element ^= 0xFF ;
    EEBlock . Password [ Loop ] = ( uint8_t ) Element ;
  } /* for */ ;
} /* DecodePassword */


static uint16_t ComputeCRC ( uint16_t  Length ,
                             uint8_t * Buffer )
{
  uint16_t Loop ;
  uint16_t ResultCRC ;
  uint16_t WorkCRC ;
  /*----------------------------------------------------------------*/
  ResultCRC = 0xFFFF ;
  for ( Loop = 0 ; Loop < Length ; Loop ++ )
  {
    WorkCRC = CRCTable [ ResultCRC & 0x0F ] ;
    ResultCRC = ( ResultCRC >> 4 ) & 0x0FFF ;
    ResultCRC ^= WorkCRC ^ CRCTable [ Buffer [ Loop ] & 0x0F ] ;
    WorkCRC = CRCTable [ ResultCRC & 0x0F ] ;
    ResultCRC = ( ResultCRC >> 4 ) & 0x0FFF ;
    ResultCRC ^= WorkCRC ^ CRCTable [ ( Buffer [ Loop ] >> 4 ) & 0x0F ] ;
  } /* for */ ;
  return ( ResultCRC ) ;
} /* ComputeCRC */


static void ReadEEBlock ( void )
{
  /*----------------------------------------------------------------*/
  eeprom_read_block ( & EEBlock , ( void * ) EEBlockAddressOffset ,
                      sizeof ( EEPROMBlockDataRecT ) ) ;
} /* ReadEEBlock */


static void WriteEEBlock ( void )
{
  uint16_t LocalCRCValue ;
  /*----------------------------------------------------------------*/
  CodePassword ( ) ;
  wdt_disable ( ) ;
  EEBlock . CRCValue = 0 ;
  LocalCRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) -
                          sizeof ( uint16_t ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  EEBlock . CRCValue = LocalCRCValue ;
  eeprom_write_block ( & EEBlock , ( void * ) EEBlockAddressOffset ,
                       sizeof ( EEPROMBlockDataRecT ) ) ;
  DecodePassword ( ) ;
  wdt_enable ( WDTO_1S ) ;
} /* WriteEEBlock */


static void EEBlockOperation ( void )
{
  uint16_t CRCValue ;
  /*----------------------------------------------------------------*/
  ReadEEBlock ( ) ;
  if ( EEBlock . EEBlockMetric != StandardSygnature )
  {
    SetDefaultEEBlock ( ) ;
    WriteEEBlock ( ) ;
    return ;
  } /* if */ ;
  CRCValue = ComputeCRC ( sizeof ( EEPROMBlockDataRecT ) ,
                          ( uint8_t * ) ( & EEBlock ) ) ;
  DecodePassword ( ) ;
  if ( CRCValue != 0 )
  {
    SetDefaultEEBlock ( ) ;
    WriteEEBlock ( ) ;
  } /* if */ ;
} /* EEBlockOperation */


static uint8_t CompareFlash ( uint8_t * Text ,
                              uint16_t  Pattern )
{
  uint8_t Ch1 ;
  uint8_t Ch2 ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Ch1 = * Text ++ ;
    Ch2 = pgm_read_byte_near ( Pattern ++ ) ;
    if ( Ch1 != Ch2 )
    {
      return ( false ) ;
    } /* if */ ;
    if ( Ch1 == 0 )
      return ( true ) ;
  } /* for */ ;
} /* CompareFlash */


static uint8_t CompareDeviceID ( uint8_t * DeviceName ,
                                 uint8_t * Pattern )
{
  uint8_t Ch1 ;
  uint8_t Ch2 ;
  /*----------------------------------------------------------------*/
  for ( ; ; )
  {
    Ch1 = * DeviceName ++ ;
    Ch2 = * Pattern ++ ;
    if ( Ch1 != Ch2 )
    {
      return ( false ) ;
    } /* if */ ;
    if ( ( Ch1 == null ) || ( Ch2 == null ) )
      return ( true ) ;
  } /* for */ ;
} /* CompareDeviceID */


static uint8_t MeDestination ( void )
{
  return ( CompareDeviceID ( UpService . DestinationDeviceName , EEBlock . MyDeviceID ) ) ;
} /* MeDestination */


static uint8_t GeneralDestination ( void )
{
  return ( CompareFlash ( UpService . DestinationDeviceName , ( uint16_t ) GeneralPoolDeviceName ) ) ;
} /* GeneralDestination */


static void AddToCyclicReplyForCommand ( void )
{
  AddToCyclic ( 'R' ) ;
  AddToCyclicTxt ( UpService . CommandText ) ;
  AddToCyclic ( '=' ) ;
} /* AddToCyclicReplyForCommand */


static void AddToCyclicReplyDevicesID ( void )
{
  AddToCyclicTxt ( UpService . SourceDeviceName ) ;
  AddToCyclic ( PartSeparator ) ;
  AddToCyclicTxt ( UpService . DestinationDeviceName ) ;
  AddToCyclic ( PartSeparator ) ;
} /* AddToCyclicReplyDevicesID */


static void SendCommandReplyOK ( void )
{
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) OK ) ;
} /* SendCommandReplyOK */


static void AddToCyclicCommandReplyFail ( void )
{
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) Fail ) ;
} /* AddToCyclicCommandReplyFail */


static void ConnectCommandExecute ( void )
{
  uint8_t Ch ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Ch = EEBlock . Password [ Loop ] ;
    if ( UpService . ParamText [ Loop ] != Ch )
      return ;
    if ( Ch == null )
      break ;
  } /* for */ ;
  ConnectContext = true ;
  SendCommandReplyOK ( ) ;
  LEDOn ( ConnectLED ) ;
} /* ConnectCommandExecute */


static void DisconnectCommandExecute ( void )
{
  ConnectContext = false ;
  SendCommandReplyOK ( ) ;
  LEDOff ( ConnectLED ) ;
} /* DisconnectCommandExecute */


static uint8_t ExecuteAssociatonCommand ( uint8_t * String ,
                                          uint8_t   AssociationTableSize ,
                                          uint16_t  AssociationTable )
{
  uint8_t Loop ;
  CommandNameAssociationRecT AssociationElement ;
  uint16_t MemAddress ;
  /*----------------------------------------------------------------*/
  MemAddress = AssociationTable ;
  for ( Loop = 0 ; Loop < AssociationTableSize ; Loop ++ )
  {
    AssociationElement . CommandService = ( CommandExecuteServiceProcT ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    AssociationElement . CommandName = ( uint8_t * ) pgm_read_word_near ( MemAddress ) ;
    MemAddress += 2 ;
    if ( CompareFlash ( String , ( uint16_t ) AssociationElement . CommandName ) )
    {
      AssociationElement . CommandService ( ) ;
      return ( true ) ;
    } /* if */ ;
  } /* for */ ;
  return ( false ) ;
} /* ExecuteAssociatonCommand */


static void HelloCommandExecute ( void )
{
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) DeviceName ) ;
} /* HelloCommandExecute */


static void GetVersionCommandExecute ( void )
{
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicFlashTxt ( ( uint16_t ) HardwareSoftwareVersion ) ;
} /* GetVersionCommandExecute */


static void GetIDCommandExecute ( void )
{
  AddToCyclicReplyDevicesID ( ) ;
  AddToCyclicReplyForCommand ( ) ;
  AddToCyclicTxt ( EEBlock . MyDeviceID ) ;
  AddToCyclic ( cr ) ;
} /* GetIDCommandExecute */


static void ReinitCommandExecute ( void )
{
  if ( ! ConnectContext )
    return ;
  EEBlockOperation ( ) ;
  SendCommandReplyOK ( ) ;
} /* ReinitCommandExecute */


static void ResetEECommandExecute ( void )
{
  if ( ! ConnectContext )
    return ;
  SetDefaultEEBlock ( ) ;
  SendCommandReplyOK ( ) ;
} /* ResetEECommandExecute */


static void SaveEECommandExecute ( void )
{
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  WriteEEBlock ( ) ;
  SendCommandReplyOK ( ) ;
} /* SaveEECommandExecute */


static void Assign ( uint8_t * SubDeviceID )
{
  uint8_t Loop ;
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  ClearBlock ( SubDeviceID , sizeof ( DeviceIDType ) ) ;
  for ( Loop = 0 ; Loop < DeviceIDSize ; Loop ++ )
  {
    Ch = UpService . ParamText [ Loop ] ;
    if ( ! Ch )
      break ;
    * SubDeviceID ++ = Ch ;
  } /* for */ ;
} /* Assign */


static void SetDeviceNameCommandExecute ( void )
{
  if ( ! ConnectContext )
    return ;
  if ( ! UpService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  Assign ( EEBlock . MyDeviceID ) ;
  SendCommandReplyOK ( ) ;
} /* SetDeviceNameCommandExecute */


static void NewPasswordCommandExecute ( void )
{
  uint8_t Ch ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! ConnectContext )
    return ;
  for ( Loop = 0 ; Loop <= PasswordSize ; Loop ++ )
  {
    Ch = UpService . ParamText [ Loop ] ;
    EEBlock . Password [ Loop ] = Ch ;
    if ( Ch == null )
      break ;
  } /* for */ ;
  SendCommandReplyOK ( ) ;
} /* NewPasswordCommandExecute */


static void ReplyCommandExecute ( void )
{
  if ( ! CyclicDataPresent ( ) )
  {
    AddToCyclicTxt ( UpService . SourceDeviceName ) ;
    AddToCyclic ( PartSeparator ) ;
    AddToCyclicTxt ( UpService . DestinationDeviceName ) ;
    AddToCyclic ( PartSeparator ) ;
    AddToCyclicFlashTxt ( ( uint16_t ) NoDataText ) ;
  } /* if */ ;
  FlushCyclic ( ) ;
} /* ReplyCommandExecute */


static uint8_t HexConv ( uint8_t Ch )
{
  /*----------------------------------------------------------------*/
  switch ( Ch )
  {
    case '0' :
      return 0 ;
    case '1' :
      return 1 ;
    case '2' :
      return 2 ;
    case '3' :
      return 3 ;
    case '4' :
      return 4 ;
    case '5' :
      return 5 ;
    case '6' :
      return 6 ;
    case '7' :
      return 7 ;
    case '8' :
      return 8 ;
    case '9' :
      return 9 ;
    case 'A' :
    case 'a' :
      return 10 ;
    case 'B' :
    case 'b' :
      return 11 ;
    case 'C' :
    case 'c' :
      return 12 ;
    case 'D' :
    case 'd' :
      return 13 ;
    case 'E' :
    case 'e' :
      return 14 ;
    case 'F' :
    case 'f' :
      return 15 ;
    default :
      return 0 ;
  } /* switch */ ;
} /* HexConv */ ;


static void DispBinCommandExecute ( void )
{
  uint8_t LocalBuffer [ 4 ] ;
  uint8_t * ChP ;
  uint8_t Ch ;
  uint8_t Inx ;
  uint8_t Data ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! UpService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  LocalBuffer [ 0 ] = 0 ;
  LocalBuffer [ 1 ] = 0 ;
  LocalBuffer [ 2 ] = 0 ;
  LocalBuffer [ 3 ] = 0 ;
  ChP = UpService . ParamText ;
  Ch = * ChP ;
  Inx = 0 ;
  Data = 0 ;
  for ( ; ; )
  {
    Ch = * ChP ++ ;
    if ( ! Ch )
    {
      LocalBuffer [ Inx ] = Data ;
      Data = 0 ;
      break ;
    } /* if */ ;
    if ( Ch == '.' )
    {
      LocalBuffer [ Inx ] = Data ;
      Data = 0 ;
      Inx ++ ;
      if ( Inx >= 4 )
        break ;
      continue ;
    } /* if */ ;
    Data = ( Data << 4 ) + HexConv ( Ch ) ;
  } /* for */ ;
  for ( Loop = 0 ; Loop < 4 ; Loop ++ )
  {
    DispBuffer [ Loop ] = LocalBuffer [ Loop ] ;
  } /* for */ ;
} /* DispBinCommandExecute */


static void DispCommandExecute ( void )
{
  uint8_t LocalBuffer [ 4 ] ;
  uint8_t * ChP ;
  uint8_t Ch ;
  uint8_t Inx ;
  uint8_t Loop ;
  /*----------------------------------------------------------------*/
  if ( ! UpService . ParamText [ 0 ] )
  {
    AddToCyclicCommandReplyFail ( ) ;
    return ;
  } /* if */ ;
  LocalBuffer [ 0 ] = 0 ;
  LocalBuffer [ 1 ] = 0 ;
  LocalBuffer [ 2 ] = 0 ;
  LocalBuffer [ 3 ] = 0 ;
  Inx = 0 ;
  ChP = UpService . ParamText ;
  Ch = * ChP ;
  if ( Ch == '+' )
  {
    ChP ++ ;
  } /* if ... */
  else
  {
    if ( Ch == '-' )
    {
      LocalBuffer [ Inx ] = ConvToSegments ( '-' ) ;
      Inx ++ ;
      ChP ++ ;
    } /* if ... */
  } /* if ... else */ ;
  for ( ; ; )
  {
    Ch = * ChP ++ ;
    if ( ! Ch )
      break ;
    if ( ( Ch == '.' ) || ( Ch == ',' ) )
    {
      if ( Inx > 0 )
        LocalBuffer [ Inx - 1 ] |= SegmentK ;
    } /* if ... */
    else
    {
      LocalBuffer [ Inx ] |= ConvToSegments ( Ch ) ;
      Inx ++ ;
      if ( Inx >= 4 )
        break ;
    } /* if ... else */ ;
  } /* for */ ;
  if ( LocalBuffer [ 3 ] )
  {
    Inx = 3 ;
  } /* if ... */
  else
  {
    if ( LocalBuffer [ 2 ] )
    {
      Inx = 2 ;
    } /* if ... */
    else
    {
      if ( LocalBuffer [ 1 ] )
      {
        Inx = 1 ;
      } /* if ... */
      else
      {
        Inx = 0 ;
      } /* if ... else */ ;
    } /* if ... else */ ;
  } /* if ... else */ ;
  for ( Loop = 0 ; Loop < 4 ; Loop ++ )
  {
    if ( Inx < 4 )
    {
      Ch = LocalBuffer [ Inx ] ;
      if ( Inx == 0 )
      {
        Inx = 10 ;
      } /* if ... */
      else
      {
        Inx -- ;
      } /* if ... else */ ;
    } /* if ... */
    else
    {
      Ch = 0 ;
    } /* if ... else */ ;
    DispBuffer [ 3 - Loop ] = Ch ;
  } /* for */ ;
} /* DispCommandExecute */


static void ExecuteCommand ( void )
{
  uint8_t FrameToMe ;
  /*----------------------------------------------------------------*/
  FrameToMe = MeDestination ( ) ;
  if ( ! FrameToMe )
  {
    FrameToMe = GeneralDestination ( ) ;
    if ( FrameToMe )
      StrCopy ( UpService . DestinationDeviceName ,
                EEBlock . MyDeviceID ) ;
  } /* if */ ;
  if ( FrameToMe )
  {
    if ( ! ExecuteAssociatonCommand ( UpService . CommandText ,
                                      CommandNameAssociationTabeSize ,
                                      ( uint16_t ) CommandNameAssociationTabe ) )
    {
    } /* if */ ;
    return ;
  } /* if */ ;
} /* ExecuteCommand */


static void ResetOperation ( void ) ;

static void DestinationIDPutCh ( uint8_t Ch )
{
  if ( UpService . DestDeviceNameIndex < DeviceIDSize )
  {
    UpService . DestinationDeviceName [ UpService . DestDeviceNameIndex ++ ] = Ch ;
  } /* if */ ;
} /* DestinationIDPutCh */


static void SourceIDPutCh ( uint8_t Ch )
{
  if ( UpService . SrcDeviceNameIndex < DeviceIDSize )
  {
    UpService . SourceDeviceName [ UpService . SrcDeviceNameIndex ++ ] = Ch ;
  } /* if */ ;
} /* SourceIDPutCh */


static void CommandIDPutCh ( uint8_t Ch )
{
  if ( UpService . CommandTextIndex < CommandBufferSize )
  {
    UpService . CommandText [ UpService . CommandTextIndex ++ ] = Ch ;
  } /* if */ ;
} /* CommandIDPutCh */


static void ParamPutCh ( uint8_t Ch )
{
  if ( UpService . ParamTextIndex < ParamBufferSize )
  {
    UpService . ParamText [ UpService . ParamTextIndex ++ ] = Ch ;
  } /* if */ ;
} /* ParamPutCh */


static void SelectParam ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch == cr )
    {
      ParamPutCh ( null ) ;
      ExecuteCommand ( ) ;
      ResetOperation ( ) ;
      return ;
    } /* if */;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    ParamPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectParam */


static void SelectCommand ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch == cr )
    {
      CommandIDPutCh ( null ) ;
      ExecuteCommand ( ) ;
      ResetOperation ( ) ;
      return ;
    } /* if */;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == ParamSeparator )
    {
      CommandIDPutCh ( null ) ;
      UpService . ParserService = SelectParam ;
      return ;
    } /* if */;
    CommandIDPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectCommand */


static void SelectSourceService ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == PartSeparator )
    {
      SourceIDPutCh ( null ) ;
      UpService . ParserService = SelectCommand ;
      return ;
    } /* if */;
    SourceIDPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectSourceService */


static void SelectDestOperation ( void )
{
  uint8_t Ch ;
  /*----------------------------------------------------------------*/
  if ( SerialDataPresent ( ) )
  {
    Ch = GetSerial ( ) ;
    if ( Ch < Space )
    {
      ResetOperation ( ) ;
      return ;
    } /* if */ ;
    if ( Ch == PartSeparator )
    {
      DestinationIDPutCh ( null ) ;
      UpService . ParserService = SelectSourceService ;
      return ;
    } /* if */;
    DestinationIDPutCh ( Ch ) ;
  } /* if */ ;
} /* SelectDestOperation */


static void ResetOperation ( void )
{
  ClearBlock ( UpService . CommandText , sizeof ( CommandBufferType ) ) ;
  ClearBlock ( UpService . ParamText , sizeof ( ParamBufferType ) ) ;
  ClearBlock ( UpService . DestinationDeviceName , sizeof ( DeviceIDType ) ) ;
  ClearBlock ( UpService . SourceDeviceName , sizeof ( DeviceIDType ) ) ;
  UpService . CommandTextIndex = 0 ;
  UpService . ParamTextIndex = 0 ;
  UpService . DestDeviceNameIndex = 0 ;
  UpService . SrcDeviceNameIndex = 0 ;
  UpService . ParserService = SelectDestOperation ;
} /* ResetOperation */


static void SoftwareInit ( void )
{
  ConnectContext = false ;
  CurrentTimeStatusState = 0 ;
  UpService . RXDRec . CycFr = 0 ;
  UpService . RXDRec . CycTo = 0 ;
  UpService . TXDRec . CycFr = 0 ;
  UpService . TXDRec . CycTo = 0 ;
  UpService . TXDRec . State = IdleState ;
  UpService . SerialLEDDelay = 0 ;
  CyclicBuffer . CycFr = 0 ;
  CyclicBuffer . CycTo = 0 ;
  TimerCounter = 0 ;
  ResetOperation ( ) ;
  TimeEvent = false ;
  DispBuffer [ 0 ] = 0 ;
  DispBuffer [ 1 ] = 0 ;
  DispBuffer [ 2 ] = 0 ;
  DispBuffer [ 3 ] = 0 ;
  DigitCounter = 0 ;
} /* SoftwareInit */


int main ( void )
{
  HardwareInit ( Speed38400bps ) ;
  EnvirInit ( ) ;
  SoftwareInit ( ) ;
  EEBlockOperation ( ) ;
  UARTReceiveMode ( ) ;
  wdt_enable ( WDTO_1S ) ;
  sei ( ) ;
  for ( ; ; )
  {
    wdt_reset ( ) ;
    if ( TimeEvent )
    {
      TimeEvent = false ;
      if ( CurrentTimeStatusState )
        LEDOn ( LifeStatusLED ) ;
      else
        LEDOff ( LifeStatusLED ) ;
      CurrentTimeStatusState = 1 - CurrentTimeStatusState ;
    } /* if */ ;
    if ( UpService . ParserService )
      UpService . ParserService ( ) ;
  } /* for */ ;
  return ( 0 ) ;
} /* main */


